import pandas as pd
import numpy as np

from sklearn.metrics import mean_absolute_error, mean_squared_error, median_absolute_error, r2_score, explained_variance_score, classification_report

class Test:
    ''' Calculates different metrics (percentage error, mean squared error, R2 ...) '''
    
    def test_percentage_error(self, error_df):
        ''' Returns a list of percentage error values '''
        
        # Create error list
        percentage_error_list = []
        
        # Iterate over rows to calculate PE
        for index, row in error_df.iterrows():
            percentage_error = (abs(row[error_df.columns[0]] - row[error_df.columns[1]]) / row[error_df.columns[0]]) * 100
            percentage_error_list.append(percentage_error)
            
        return percentage_error_list
    
    def test_error_values(self, error_df):
        ''' Return a list of error metrics '''
        
        mepe = round(error_df['error(%)'].median(), 3)
        mpe = round(error_df['error(%)'].mean(), 3)

        meae = round(median_absolute_error(error_df['true_values'], error_df['predictions']), 3)
        mae = round(mean_absolute_error(error_df['true_values'], error_df['predictions']), 3)
        mse = round(mean_squared_error(error_df['true_values'], error_df['predictions']), 3)
        rmse = round(np.sqrt(mse), 3)
        normalized_rmse = rmse/(error_df['true_values'].max()-error_df['true_values'].min())

        std = round(error_df['true_values'].std(), 3)

        error_list = [mepe, mpe, meae, mae, mse, rmse, normalized_rmse, std]

        return error_list
    
    def test_variability_values(self, error_df, X_test):
        ''' Returns a list of variability values '''
    
        # R2 Score
        r2 = r2_score(error_df['true_values'], error_df['predictions'])
        
        # Adj R2 ---> 1-(1-R2)*(n-1)/(n-p)
        adj_r2 = 1-((1-r2) * ((X_test.shape[0]-1) / (X_test.shape[0] - X_test.shape[1]-1)))
        
        # Explained Varience Score
        evs = explained_variance_score(error_df['true_values'], error_df['predictions'])

        variability_values_list = [r2, adj_r2, evs]

        return variability_values_list
    
    def test_classification_report(self, y_true, y_pred):
        ''' Returns a dictionary of classification report results'''
        classification_report_dictionary = classification_report(y_true, y_pred, output_dict=True)
        report_list = list(classification_report_dictionary.keys())
        report_classes_list = report_list[:-3]
        accuracy = classification_report_dictionary['accuracy']

        for class_ in report_classes_list: classification_report_dictionary[class_]['accuracy'] = accuracy

        return classification_report_dictionary


class Check:
    
    def check_if_size_is_smaller_than_two(self, size):
    
        if size <= 2:
            size_is_smaller_than_two = True

        else:
            size_is_smaller_than_two = False

        return size_is_smaller_than_two

    def check_if_element_is_in_data(self, element, data):

        if element in data:
            element_is_in_data = True
            
        else:
            element_is_in_data = False
            
        return element_is_in_data
    
    def check_if_elements_are_equal(self, first_element, second_element):
        
        if first_element == second_element:
            elements_are_equal = True
            
        else:
            elements_are_equal = False
        
        return elements_are_equal
        
    def check_if_there_are_multiple_words(self, string):
        
        word_count = len(string.split(' '))
        
        if word_count > 1:
            there_are_multiple_words = True
        
        else:
            there_are_multiple_words = False
            
        return there_are_multiple_words
        
    def check_if_there_are_more_than_two_elements(self, data):
        
        element_count = len(data)
        
        if element_count > 2:
            
            there_are_more_than_two_elements = True
        
        else:
            
            there_are_more_than_two_elements = False
            
        return there_are_more_than_two_elements
    
    def check_if_there_are_two_elements(self, data):
        
        element_count = len(data)
        
        if element_count == 2:
            
            there_are_two_elements = True
        
        else:
            
            there_are_two_elements = False
            
        return there_are_two_elements
    
    def check_if_there_is_one_element(self, data):
        
        element_count = len(data)
        
        if element_count == 1:
            
            there_is_one_element = True
        
        else:
            
            there_is_one_element = False
            
        return there_is_one_element
    
    def check_if_indexes_are_consecutive(self, first_index, second_index):

        if first_index + 1 == second_index:

            indexes_are_consecutive = True

        else:

            indexes_are_consecutive = False

        return indexes_are_consecutive
    
    def check_if_order_is_correct(self, first_index, second_index):
    
        if first_index < second_index:
            order_is_correct = True

        else:
            order_is_correct = False
            
        return order_is_correct

class BaseFormat:
    
    def format_passenger_string(self, passenger_string):
        
        formatted_passenger_string = passenger_string.split(',')[0]
        
        return formatted_passenger_string
    
    def format_date_to_string(self, date_object):

        formatted_date_string = date_object.strftime("%d-%m-%Y")
        
        return formatted_date_string

    def format_string_with_num(self, string_with_num):
        
        """ 
        Reformat algorithm_string to get algorithm name
        
        """
        
        digit_num = len([letter for letter in string_with_num if letter.isdigit()])
        string_without_num = string_with_num[:-digit_num]
        
        return string_without_num


import string
from math import ceil

class Generate:

    def generate_numerical_ranges(self, numeric_variable, thresholds_list):
        ''' Create a list of ranges for numeric variable you want to categorize

            Parameters
            ----------
            numeric_variable: Series
                        Column of a DataFrame that contains numerical values that you want to categorize
            thresholds_list: List
                        List of the thresholds 
        '''
        numerical_ranges_list = []
        start = int(numeric_variable.min()-1)
        final = int(numeric_variable.max()*1.5)
        for threshold in thresholds_list:
            numerical_ranges_list.append(range(start, threshold))
            start = threshold

        numerical_ranges_list.append(range(threshold, final))

        return numerical_ranges_list

    def generate_categorical_ranges(self, interval,start,finish):
        ''' Create a list of ranges for categorcial variable you want to categorzie

            Parameters
            ----------
            interval: int
                Count of intervals to create
            start: int -- float
                minnimum value
            finish: int -- float
                maximum value
        '''
        categorical_ranges_list = []
        start = int(start)
        finish = int(finish)
        diff = int((finish - start) / interval)

        for i in range(interval):
            if interval-i != 1:
                categorical_ranges_list.append(range(start, start+ diff))
            else:
                categorical_ranges_list.append(range(start, ceil(1.5 * (start+diff))))

            start += diff

        return categorical_ranges_list
    


import pandas as pd
import string
from math import ceil
from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import train_test_split

from automl.test import Test

from tensorflow.keras.layers import Input, Dense


class Format:
    
    def format_data_string(self, data_string):
        """
        Reformat data_string to store data
        """

        # Retrieve indexes of digits inside the data string
        indexes = [index for index, letter in enumerate(data_string) if letter.isdigit()]

        for index, index_value in enumerate(indexes):

            # Assign the value on the previous index
            previous_index_value = indexes[index - 1]

            if index > 0:

                # Compare index values that are in order to find the cut between data and algorithm strings
                if previous_index_value + 1 != index_value:

                    # Find the cut then break! 
                    # Otherwise would be retrieving the entire string or get and error

                    end_index_of_indexes = indexes.index(index_value)- 1
                    end_index_of_data_string = indexes[end_index_of_indexes] + 1
                    break

        return data_string[:end_index_of_data_string]
    
    
    def format_classification_report(self, classification_report_dictionary):
        """
        Reformat classification report dictionary to store result in a dataframe
        """
        del classification_report_dictionary['accuracy']
        del classification_report_dictionary['macro avg']
        del classification_report_dictionary['weighted avg']

        classification_report_df = pd.DataFrame.from_dict(classification_report_dictionary).transpose()
        classification_report_df = classification_report_df.reset_index()
        classification_report_df = classification_report_df.rename(columns={'index': 'class'})

        return classification_report_df
    
    
    def format_input_layer(self, layers, X_train):

        n_features = X_train.shape[1]
        input_layer = Input(shape=(n_features,))

        if type(input_layer) == type(layers[0]):

            units = [lyr.input_shape[1] for lyr in layers[1:]]
            units.append(1)

            layers = [Dense(unit, activation ='relu') for unit in units]
            del layers[0]

            layers.insert(0, input_layer)

        else:
            layers.insert(0, input_layer)

        return layers

    
    def format_prophet_data(self, data):
        
        data = data.reset_index()
        data.columns = ['ds','y']
        
        return data
    
    def format_prophet_index(self, train, test):
        
        train = train.set_index('ds')
        test = test.set_index('ds')
        
        return train, test
    
    def format_prophet_predictions(self, predictions):
        
        predictions = predictions[['ds', 'yhat']]
        predictions = predictions.set_index('ds')
        
        return predictions
    
    def format_models_list(self, model):
        
        return list(model.values())
    
class Assemble(Test, Format):
    ''' Create dataframse with test values '''
    
    def assemble_test_tables(self, regression_test_dict):
        ''' Create a dataframe with y, prediction and percentage error values and add it to dictionary'''
        
        test_tables_dict = {}
        for key, value in regression_test_dict['predictions'].items():

            error_df = pd.DataFrame()
            data_name = self.format_data_string(key)
            y_true = regression_test_dict['y_test'][data_name]
            error_df['true_values'] = y_true
            error_df['predictions'] = value

            # Retrieve percentage error values for each model
            error_df['error(%)'] = self.test_percentage_error(error_df)

            test_tables_dict[key] = error_df
        return test_tables_dict
        
    def assemble_error_values(self, regression_test_dict):
        ''' Returns a dataframe with error metrics for each model '''
        
        error_lists = []
        model_name_list = []
        for key, value in regression_test_dict['test_tables'].items():

            error_list = self.test_error_values(value)
            error_lists.append(error_list)
            model_name_list.append(key)

        error_values_df = pd.DataFrame(error_lists, index=model_name_list, columns = ['MEPE','MPE','MEAE','MAE','MSE','RMSE', 'NRMSE', 'STD'])
        error_values_df.index.names = ['ALGORITHM']
 
        return error_values_df
    
    def assemble_variability_values(self, regression_test_dict):
        ''' Returns a dataframe with variability values for each model'''
        
        variability_values_lists = []   
        model_name_list = []
        for key, value in regression_test_dict['test_tables'].items():
            
            data_name = self.format_data_string(key)
            variability_values_list = self.test_variability_values(value, regression_test_dict['X_test'][data_name])  
            variability_values_lists.append(variability_values_list)
            model_name_list.append(key)

        variability_values_df = pd.DataFrame(variability_values_lists, index=model_name_list, columns = ['R2', 'R2^', 'EVS'])
        variability_values_df.index.names = ['ALGORITHM']
        return variability_values_df
    
    def assemble_classification_report(self, classification_test_dict):
        ''' Returns a dataframe with classification report values for each model '''
        
        classification_df= pd.DataFrame()

        for key, value in classification_test_dict['predictions'].items():
            data_name = self.format_data_string(key)
            y_true = classification_test_dict['y_test'][data_name]
            y_pred = value
            
            classification_report_dictionary = self.test_classification_report(y_true, y_pred)
            classification_report_df = self.format_classification_report(classification_report_dictionary)
            classification_report_df['ALGORITHM'] = key
            classification_report_df.set_index('ALGORITHM', inplace=True)
            classification_df = pd.concat([classification_df, classification_report_df])
        classification_df.columns = classification_df.columns.str.upper()
        return classification_df
    
class Categorize:
    
    def __init__(self):
        self.letters = string.ascii_letters
    
    def categorize_numerical_variable(self, df_series, numerical_ranges_list):
        ''' Returns a dictionary of categories with corresponding ranges 
        
            Parameters
            ----------
            df_series: Series
                    Numerical variable series
            numerical_ranges_list: List
                    List of ranges for numerical values
        '''
        
        letters = self.letters[:len(numerical_ranges_list)]
        range_dict = dict(zip(letters, numerical_ranges_list))
        category_dict = {}

        for key, value in range_dict.items():
            for data in  df_series.values:
                if int(data) in value:
                    category_dict[data] = key

        return category_dict
    
    def categorize_categorical_variable(self, grouped_data, categorical_ranges_list):
        ''' Returns a dictionary of categories with corresponding ranges 

            Parameters
            ----------
            grouped_data: Series
                    Series acquired after using groupby method
            categorical_ranges_list: List
                    List of ranges for categorical values
        '''

        letters = string.ascii_letters[:len(categorical_ranges_list)]
        range_dict = dict(zip(letters, categorical_ranges_list))
        category_dict = {}
        for name, data in enumerate(grouped_data):
            for key, value in range_dict.items():
                if int(data) in value:
                    category_dict[grouped_data.index[name]] = key

        return category_dict

class Transform:
    
    def transform_data(self, data, feature_range = (0,1)):
        
        scaler = MinMaxScaler(feature_range = feature_range)
        data_scaled = scaler.fit_transform(data)
        return data_scaled, scaler
        
    def transform_test_data(self, X_train, X_test):    
        
        scaler = MinMaxScaler()
        X_train = scaler.fit_transform(X_train)
        X_test = scaler.transform(X_test)
        
        return X_train, X_test
    
    
class Preprocess(Transform):
    
    def preprocess_test_data(self, data, dependent_variable):
        

        X = data.drop(dependent_variable, axis=1)
        y = data[dependent_variable]
        
        X = pd.get_dummies(X, drop_first=True)

        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)
        X_train, X_test = self.transform_test_data(X_train, X_test)


        return  X_train, X_test, y_train, y_test
        
    def preprocess_data(self, data, dependent_variable):
        

        X = data.drop(dependent_variable, axis=1)
        y = data[dependent_variable]
        
        X = pd.get_dummies(X, drop_first=True)
        
        X, scaler = self.transform_data(X)

        return X, y, scaler
    
    def preprocess_time_series_data(self, data, test_size):
        
        train, test = model_selection.train_test_split(data, test_size=test_size)
        
        return train, test


from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from sklearn.cross_decomposition import PLSRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.svm import SVR

from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC

from xgboost import XGBRegressor

from sklearn.feature_selection import RFE
from sklearn.feature_selection import SequentialFeatureSelector
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import SelectFromModel

from sklearn.decomposition import PCA
from sklearn.decomposition import SparsePCA
from sklearn.decomposition import KernelPCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis


from sklearn.model_selection import train_test_split

import tensorflow as tf
from tensorflow.keras.layers import Dense, Input
from tensorflow.keras import Sequential


from automl.process import Format, Preprocess
from base.utils.servant import BaseFormat

import pmdarima as pm
from pmdarima import model_selection
from prophet import Prophet

import gc

class Create:
    
    def create_regression_model(self, algorithm, parameters = {}):
        """
        Create a regression model

        Parameters
        ----------


        algorithm : {"LR", "PLS", "RFR", "SVR", "PR"}
            algorithm abbreviation

            LR: LinearRegression
            PLS: PartialLeastSquaresRegressor
            RFR: RandomForestRegressor
            SVR: SupportVectorRegressor
            PR: PolinomialRegressor

            Note: Using "PR" would return a tuple including the regressor and the polynomial features object. Later, X_test should be transformed using 
            polinomial features object and predictions must be done on the X_test_polynomial object.

            X_test_polynomial = polynomial.transform(X_test)
            y_pred = regressor.predict(X_test_polynomial)

        parameters: dictionary
            Various parameters that could be defined in different choice of algorithms
        """
        """
        if algorithm == 'PR':

            regressor = LinearRegression()

            polynomial = PolynomialFeatures(degree = parameters['degree'])
            X_polynomial = polynomial.fit_transform(data[0])

            regressor.fit(X_polynomial, data[1])

            return (regressor, polynomial)
         """

        if algorithm == 'LR' : regressor = LinearRegression(**parameters)

        if algorithm == 'PLS': regressor = PLSRegression(**parameters)

        if algorithm == 'RFR' : regressor = RandomForestRegressor(**parameters)

        if algorithm == 'SVR' : regressor = SVR(**parameters)
        
        if algorithm == 'GBR' : regressor = GradientBoostingRegressor(**parameters)
        
        if algorithm == 'XGB' : regressor = XGBRegressor(**parameters)


        return regressor
    
    def create_classification_model(self, algorithm, parameters = {}):
        """
        Create a classification model

        Parameters
        ----------
    
        algorithm : {"LR", "DTC", "RFC", "SVC", "KNN", "LDA", "QDA"}
            algorithm abbreviation

            LR: LogisticRegression
            DTC: DecisionTreeClassifier
            RFC: RandomForestClassifier
            SVC: SupportVectorClassifier
            KNN: KNeighborClassifier
            LDA: LinearDisriminantAnalysis
            QDA: QuadriticDiscriminantAnalysis
            

        parameters: dictionary
            Various parameters that could be defined in different choice of algorithms
        """

        if algorithm == 'LR' : classifier = LogisticRegression(**parameters)

        if algorithm == 'DTC': classifier = DecisionTreeClassifier(**parameters)

        if algorithm == 'RFC' : classifier = RandomForestClassifier(**parameters)

        if algorithm == 'SVC' : classifier = SVC(**parameters)
        
        if algorithm == 'KNN' : classifier = KNeighborsClassifier(**parameters)
        
        if algorithm == 'LDA':model = LinearDiscriminantAnalysis(**parameters)

        if algorithm == 'QDA':model = QuadraticDiscriminantAnalysis(**parameters)


        return classifier
    
    def create_feature_selection_model(self, estimator_model, selector_algorithm, selector_parameters = {}):
        """
       Create a feature selection model

       Parameters
       ----------
        parameters: dictionary
                Various parameters that could be defined in different choice of algorithms

        algorithm:{'SFM','RFE','SKB','SFS'}
             algorithm abbreviation

        SFM: SelectFromModel
        RFE: RecursiveFeatureElimination
        SKB: SelectKBest
        SFS: SequentialFeatureSelection


        """
        
        if selector_algorithm == 'SFM' : feature_selector = SelectFromModel(estimator = estimator_model, **selector_parameters)

        if selector_algorithm == 'RFE': feature_selector = RFE(estimator = estimator_model, **selector_parameters)

        if selector_algorithm == 'SKB': feature_selector = SelectKBest(**selector_parameters)

        if selector_algorithm == 'SFS': feature_selector = SequentialFeatureSelector(estimator = estimator_model, **selector_parameters)
        
        return feature_selector
    
    def create_dimensionality_reduction_model(self,algorithm,parameters={}):
        """
        Create a dimensionality reduction model

        Parameters
        ----------
        parameters: dictionary
                Various parameters that could be defined in different choice of algorithms

        algorithm:{'PCA','SPCA','LDA'}
             algorithm abbreviation

        PCA: PrincipalComponentAnalysis
        SPCA: SparsePCA
        KPCA: KernelDiscriminantAnlaysis


        """

        if algorithm == 'PCA': model = PCA(**parameters)

        if algorithm == 'SPCA': model = SparsePCA(**parameters)

        if algorithm == 'KPCA':model = KernelPCA(**parameters)
        
        if algorithm == 'AE': model = self.create_autoencoder_model(parameters['encoder_layers'], parameters['decoder_layers'], parameters['compile_parameters'])

        return model
    
    def create_deep_learning_model(self, layers, compile_parameters={}):
        """
        Create and compile a deep learning model
        
        Parameters
        ----------
        layers: list
            List of neural network layers
            
        compile_parameters: dictionary
            Parameters for compile method in dictionary format
        """
                
        network = Sequential(layers)
        network.compile(**compile_parameters)

        return network
    
    def create_autoencoder_model(self, encoder_layers, decoder_layers, compile_parameters={}):
        """
        Create and compile a autoencoder model. Returns Autoencoder and Encoder in tuple format
        
        Parameters
        ----------
        encoder_layers: list
            List of encoder layers
            
        decoder_layers: list
            List of decoder layers
            
        compile_parameters: dictionary
            Parameters for compile method in dictionary format
        """
        encoder = Sequential(encoder_layers)
        decoder = Sequential(decoder_layers)
        autoencoder = Sequential([encoder,decoder])
        autoencoder.compile(**compile_parameters)
        
        return (autoencoder, encoder)
    
    def create_time_series_model(self, algorithm, data, parameters={}):
        
        if algorithm == 'AAR' : ts_model = pm.auto_arima(data, **parameters)

        if algorithm == 'PRH': ts_model = Prophet(**parameters)
        
        return ts_model
    

class Include(Create, Format, Preprocess, BaseFormat):

    def include_feature_selection(self, model, key, model_object, X_train, X_test, y_train):
        """
        Apply feature selection on train and test values
        """
        selector_model = model[1] 
        selector_name = selector_model['feature_selection'][0]
        selector_algorithm = self.format_algorithm_string(selector_name)
        selector_parameters = selector_model['feature_selection'][1]

        print(f'Selecting features with {selector_name} for {key}')
        feature_selector = self.create_feature_selection_model(model_object, selector_algorithm, selector_parameters)

        X_train_selected = feature_selector.fit_transform(X_train, y_train)
        X_test_selected = feature_selector.transform(X_test)
        
        return X_train_selected, X_test_selected, selector_name
    
    def include_dimensionality_reduction(self, model, key, X_train, X_test, y_train):
        """
        Apply dimensionality reduction on train and test values
        """
        dimensionality_reduction_model = model[1] 
        dimensionality_reduction_name = dimensionality_reduction_model['dimensionality_reduction'][0]
        dimensionality_reduction_algorithm = self.format_algorithm_string(dimensionality_reduction_name)
        dimensionality_reduction_parameters = dimensionality_reduction_model['dimensionality_reduction'][1]
        

        print(f'Reducing dimensions with {dimensionality_reduction_name} for {key}')
        
        dimensionality_reducer = self.create_dimensionality_reduction_model(dimensionality_reduction_algorithm, dimensionality_reduction_parameters)
        
        if dimensionality_reduction_algorithm == 'AE':
            autoencoder = dimensionality_reducer[0]
            encoder = dimensionality_reducer[1]
            autoencoder.fit(X_train, X_train, **dimensionality_reduction_parameters['fit_parameters'])
            X_train_reduced = encoder.predict(X_train)
            X_test_reduced = encoder.predict(X_test)
            
            return X_train_reduced, X_test_reduced, dimensionality_reduction_name
            
        X_train_reduced = dimensionality_reducer.fit_transform(X_train)
        X_test_reduced = dimensionality_reducer.transform(X_test)
        
        return X_train_reduced, X_test_reduced, dimensionality_reduction_name
    
    
class Build(Include):
    
    def __init__(self, test_dict, feature_selection=False, dimensionality_reduction=False):
        
        self.test_dict = test_dict
        self.test_dict['predictions'] = {}
        self.test_dict['models'] = {}
        self.test_dict['X_test'] = {}
        self.test_dict['y_test'] = {}
        self.test_dict['results'] = {}
        self.feature_selection = feature_selection
        self.dimensionality_reduction = dimensionality_reduction
        
    def build_regression_models(self, models_list, dependent_variable):
        """
        Build multiple regression models on multiple data.
        
        Parameters
        ----------
        models_list: list
            List of regression models and parameters.
            
        dependent_variable: string
            Name of the variable that model is going to predict
        """      
        for key, data in self.test_dict['data'].items():
            for model in models_list:
                
                X_train, X_test, y_train, y_test = self.preprocess_test_data(data, dependent_variable)
                
                model_name, parameters = self.format_models_list(model[0])
                algorithm = self.format_algorithm_string(model_name)
                
                # Dict names
                label = key+model_name
                
                regressor = self.create_regression_model(algorithm , parameters)
                
                if self.feature_selection: 
                    X_train_selected, X_test_selected, selector_name = self.include_feature_selection(model, key, regressor, X_train, X_test, y_train)
                    X_train = X_train_selected
                    X_test = X_test_selected
                    label += selector_name

                if self.dimensionality_reduction:
                    X_train_reduced, X_test_reduced, reducer_name = self.include_dimensionality_reduction(model, key, X_train, X_test, y_train)
                    X_train = X_train_reduced
                    X_test = X_test_reduced
                    label += reducer_name

                print(f'Training regression model {model_name} for {key}')
                regressor.fit(X_train, y_train)
                print(f'Training done!')
                predictions = regressor.predict(X_test)
                print()
                
                self.test_dict['models'][label] = regressor
                self.test_dict['predictions'][label] = predictions
                self.test_dict['X_test'][key] = X_test
                self.test_dict['y_test'][key] = y_test
        
        
    def build_classification_models(self, models_list, dependent_variable):
        """
        Build multiple classification models on multiple data.
        
        Parameters
        ----------
        models_list: list
            List of classification models and parameters.
            
        dependent_variable: string
            Name of the variable that model is going to predict
        """       
        for key, data in self.test_dict['data'].items():
            for model in models_list:

                X_train, X_test, y_train, y_test = self.preprocess_test_data(data, dependent_variable)

                model_name, parameters = self.format_models_list(model[0])
                algorithm = self.format_algorithm_string(model_name)
                
                # Dict names
                label = key+model_name
                
                classifier = self.create_classification_model(algorithm , parameters)
                
                if self.feature_selection: 
                    X_train_selected, X_test_selected, selector_name = self.include_feature_selection(model, key, classifier, X_train, X_test, y_train)
                    X_train = X_train_selected
                    X_test = X_test_selected
                    label += selector_name
                    
                if self.dimensionality_reduction:
                    X_train_reduced, X_test_reduced, reducer_name = self.include_dimensionality_reduction(model, key, X_train, X_test, y_train)
                    X_train = X_train_reduced
                    X_test = X_test_reduced
                    label += reducer_name

                print(f'Training classification model {model_name} for {key}')
                classifier.fit(X_train, y_train)
                print(f'Training done!')
                predictions = classifier.predict(X_test)
                print()

                self.test_dict['models'][label] = classifier
                self.test_dict['predictions'][label] = predictions
                self.test_dict['X_test'][key] = X_test
                self.test_dict['y_test'][key] = y_test
                
    def build_deep_learning_models(self, models_list, dependent_variable, compile_parameters={}, fit_parameters={}):
        """
        Build multiple regression models on multiple data.
        
        Parameters
        ----------
        models_list: list
            List of deep learning models and parameters.
            
        dependent_variable: string
            Name of the variable that model is going to predict
        
        compile_parameters: dictionary
            Parameters that are going to be used for compiling the neural network.
        
        fit_parameters: dictionary
            Parameters that are going to be used for training the neural network
        """   
        for key, data in self.test_dict['data'].items():
            
        
            for model in models_list:
                
                X_train, X_test, y_train, y_test = self.preprocess_test_data(data, dependent_variable)
                
                model_name, layers, compile_parameters, fit_parameters = self.format_models_list(model[0])
                kind = self.format_algorithm_string(model_name)
                
                # Dict names
                label = key+model_name
                
                #Create NN 
                layers = self.format_input_layer(layers, X_train)
                nn = self.create_deep_learning_model(layers, compile_parameters)
                                
                if self.feature_selection: 
                    X_train_selected, X_test_selected, selector_name = self.include_feature_selection(model, key, nn, X_train, X_test, y_train)
                    X_train = X_train_selected
                    X_test = X_test_selected
                    label += selector_name

                if self.dimensionality_reduction:
                    X_train_reduced, X_test_reduced, reducer_name = self.include_dimensionality_reduction(model, key, X_train, X_test, y_train)
                    X_train = X_train_reduced
                    X_test = X_test_reduced
                    label += reducer_name
                
                print(f'Training deeplearning model {model_name} for {key}')
                result = nn.fit(X_train, y_train, **fit_parameters)
                print(f'Training done!')
                predictions = nn.predict(X_test)
                print()

                self.test_dict['models'][label] = nn
                self.test_dict['predictions'][label] = predictions
                self.test_dict['X_test'][key] = X_test
                self.test_dict['y_test'][key] = y_test
                self.test_dict['results'][label] = result
        
             
    def build_time_series_models(self, models_list):
        """
        Build multiple time series models on multiple data.
        
        Parameters
        ----------
        models_list: list
            List of time series models and parameters.
        """   
        for key, data in self.test_dict['data'].items():
            for model in models_list:

                model_name, layers, parameters, test_size = self.format_models_list(model[0])
                algorithm = self.format_algorithm_string(model_name) 
                
                #Change columns names if algorithm is Prophet
                if algorithm == 'PRH':
                    data = self.format_prophet_data(data)
                    
                train, test= self.preprocess_time_series_data(data, test_size)
    
                # Dict names
                label = key+model_name
                
                print(f'Creating time series model {model_name} for {key}')
                ts_model = self.create_time_series_model(algorithm, data , parameters)
                print('Model created!')
                ts_model.fit(train)
                
                if algorithm == 'PRH':
                    #Drop y values from test dataframe to make future dataframe for Prophet.
                    predictions = ts_model.predict(test.drop('y', axis=1))
                    predictions = self.format_prophet_predictions
                    
                    #Format the test and train tables so index is datetime column
                    train, test = self.format_prophet_index(train, test)
                
                else:
                    predictions = ts_model.predict(n_periods = test.shape[0])

                self.test_dict['models'][label] = ts_model
                self.test_dict['predictions'][label] = predictions
                #Because there is no X - y split in time series, train is labeled as X_test and test is labeled as y_test.
                self.test_dict['X_test'][key] = train
                self.test_dict['y_test'][key] = test

